import numpy as np
from scipy.stats import rv_discrete
import matplotlib.pyplot as plt
from itertools import product
from probability_operations.basic_probability_functions import to_rv_discrete, to_non_zero_rv, pointwise_min_cdfs, construct_rv_total_prob, mgf
from probability_operations.convolution import fast_sum_multiple_rvs


class JointRvDiscrete:
    """Class for representing and operating on a joint discrete random variable."""

    def __init__(self, joint_pmf, x_values, y_values):
        """
        Initialize the joint random variable with its joint PMF and possible values.

        Parameters:
            joint_pmf (list[list[float]]): Joint PMF as a 2D list.
            x_values (np.array): Possible values for the X variable.
            y_values (np.array): Possible values for the Y variable.
        """
        self.joint_pmf = np.array(joint_pmf)[::-1]  # Reverse the rows to align with graphical representation
        self.x_values = x_values
        self.y_values = y_values
        self.marginal_X = rv_discrete(values=(x_values, np.sum(self.joint_pmf, axis=0)))
        self.marginal_Y = rv_discrete(values=(y_values, np.sum(self.joint_pmf, axis=1)))

    def pmf(self, x, y):
        """
        Probability mass function for the joint distribution at specific (x, y).

        Parameters:
            x (int|float): The value for X.
            y (int|float): The value for Y.

        Returns:
            float: The PMF value at (x, y).
        """
        i = np.where(self.x_values == x)[0]
        j = np.where(self.y_values == y)[0]
        if len(i) == 0 or len(j) == 0:
            return 0
        return self.joint_pmf[j[0], i[0]]

    def marginal_probabilities(self):
        """
        Computes marginal probabilities for X and Y.

        Returns:
            tuple: Marginal probabilities for X and Y as dictionaries.
        """
        marginal_probs_X = {x: self.marginal_X.pmf(x) for x in self.x_values}
        marginal_probs_Y = {y: self.marginal_Y.pmf(y) for y in self.y_values}
        return marginal_probs_X, marginal_probs_Y

    def plot_joint_pmf(self, x_label='X values', y_label='Y values', title='Joint PMF'):
        """Plot the joint PMF using matplotlib."""
        fig, ax = plt.subplots()
        ax.set_title(title)
        cax = ax.matshow(self.joint_pmf[::-1], cmap='Blues')  # Reverse the rows for plotting
        ax.set_xticks(np.arange(len(self.x_values)))
        ax.set_yticks(np.arange(len(self.y_values)))
        ax.set_xticklabels(self.x_values)
        ax.set_yticklabels(reversed(self.y_values))  # Reverse the y_values for labels
        ax.xaxis.set_ticks_position('bottom')
        ax.set_xlabel(x_label)
        ax.set_ylabel(y_label)
        cbar = fig.colorbar(cax)
        cbar.set_label('Probability')

        # Annotate cells in the matrix with their probability values
        for i in range(len(self.y_values)):
            for j in range(len(self.x_values)):
                c = self.joint_pmf[-(i+1), j]  # Access the reversed row
                ax.text(j, i, '{:.4f}'.format(c), va='center', ha='center')
        plt.show()

    def pretty_print_joint_pmf(self):
        """Print the joint PMF in a readable tabular format with lines separating rows and columns."""
        
        # Define a function to create a horizontal line
        def horizontal_line(length):
            return '-' * length

        # Print footer with Marginal_X
        footer = ["C_1"] + ["{:.4f}".format(np.sum(self.joint_pmf[:, i])) for i in range(len(self.x_values))]
        footer_line = "\t".join(footer)
        print(footer_line)

        # Print horizontal line after the footer
        print(horizontal_line(len(footer_line)))

        # Print the PMF values
        for i, y in enumerate(reversed(self.y_values)):  # Reverse the y_values here
            row = [str(y)]
            for j, x in enumerate(self.x_values):
                row.append("{:.4f}".format(self.joint_pmf[-(i+1), j]))  # Access the reversed row
            row.append("|")  # Vertical line separator
            row.append("{:.4f}".format(np.sum(self.joint_pmf[-(i+1), :])))  # Access the reversed row
            print("\t".join(row))

        # Print header with x_values and Marginal_Y
        header = [" "] + [str(x) for x in self.x_values] + ["|", "C_2"]
        header_line = "\t".join(header)
        
        # Print horizontal line before the header
        print(horizontal_line(len(footer_line)))

        # Print the header
        print(header_line)


def compute_ground_truth_DFP(joint_dist, D):
    """
    Compute the ground truth deadline-failure probability (DFP) based on a given joint distribution.

    Parameters:
    joint_dist : JointRvDiscrete
        The joint probability mass function as an object of JointRvDiscrete.
    D : Relative deadline

    Returns:
    numeric
        The computed ground truth DFP value.
    """
    return sum(joint_dist.pmf(x, y)
               for x in joint_dist.x_values 
               for y in joint_dist.y_values 
               if x + y > D)

def compute_independence_assuming_DFP(response_time_dist, D):
    """
    Compute the ground truth deadline-failure probability (DFP) based on a given joint distribution.

    Parameters:
    response_time_dist : scipy.stats.rv_discrete
        An object representing a discrete random variable for response time.
    D : Relative deadline

    Returns:
    numeric
        The computed ground truth DFP value.
    """
    return sum(prob for (x, prob) in zip(response_time_dist.xk, response_time_dist.pk) if x > D)

def print_RV(RV):
    """
    Print the values with non-zero probabilities and their corresponding probabilities.

    Parameters:
    RV : scipy.stats.rv_discrete
        An object representing a discrete random variable.

    Returns:
    None
    """
    xk, pk = RV.xk, RV.pk
    for x, p in zip(xk, pk):
        if p != 0:
            print(f"Execution time: {x}, Probability: {p}")


def generate_partitions(S):
    """Generate all partitions of the set S."""
    if not S:
        yield []
        return
    first = next(iter(S))
    rest = S - {first}
    for partition in generate_partitions(rest):
        # Add first to each existing subset and yield
        for i in range(len(partition)):
            yield partition[:i] + [[first] + partition[i]] + partition[i+1:]
        # Create new subset that only contains first
        yield [[first]] + partition

def probability_of_event(event, joint_dist):
    """
    Compute the probability of an event given a joint distribution.

    Parameters:
    - event: set of tuples
        A set of (x, y) tuples representing the event in the sample space.
    - joint_dist: JointRvDiscrete object
        Object representing the joint distribution.

    Returns:
    - float
        The probability of the given event occurring.
    """
    prob = 0  # Initialize the probability of the event to zero
    for (x, y) in event:  # Loop through all (x, y) pairs in the event
        prob += joint_dist.pmf(x, y)  # Add the probability of each (x, y) pair to the event's probability
    return prob  # Return the computed probability of the event


def conditional_probability_cost_cost_other_event(task_id, cost_J, cost_J_other, event, joint_dist):
    """
    Compute the conditional probability P((cost_J, cost_J_other) | event).

    Parameters:
    - task_id: int
        The task ID, which should be either 1 or 2. Determines the role of cost_J and cost_J_other.
    - cost_J: numeric
        The cost value for task J.
    - cost_J_other: numeric
        The cost value for the other task.
    - event: set of tuples
        A set of (x, y) tuples representing the event in the sample space.
    - joint_dist: JointRvDiscrete object
        Object representing the joint distribution.

    Returns:
    - float
        The conditional probability of the given cost_J and cost_J_other occurring given the event.
    """
    
    # Determine which cost is x and which is y based on task_id
    if task_id == 1:
        x, y = cost_J, cost_J_other
    elif task_id == 2:
        x, y = cost_J_other, cost_J
    else:
        raise ValueError("Invalid task_id; must be either 1 or 2.")
        
    # Compute the probability of the event
    prob_event = probability_of_event(event, joint_dist)

    # Compute the conditional probability
    if (x, y) not in event:  # If (x, y) is not part of the event, return 0
        return 0
    else:  # Else, compute and return the conditional probability
        return joint_dist.pmf(x, y) / prob_event if prob_event > 0 else 0  # Safeguard against division by zero


def conditional_probability_cost_event(task_id, cost_J, event, joint_dist):
    """
    Compute the conditional probability that a task has a specific cost_J given an event.
    
    Parameters:
    - task_id: int
        Task ID that should be either 1 or 2. Determines whether cost_J applies to the first or second task.
    - cost_J: numeric
        The cost value for the task in question.
    - event: set of tuples
        A set of (x, y) tuples representing the event in the sample space.
    - joint_dist: JointRvDiscrete object
        Object representing the joint distribution.
        
    Returns:
    - conditional_prob: float
        The conditional probability P(cost_J | event).
    """

    # Compute the probability of the event
    prob_event = probability_of_event(event, joint_dist)

    # Initialize the probability of the intersection of cost_J and the event
    prob_intersection = 0
    
    # Compute the probability of the intersection of cost_J and the event
    for (x, y) in event:
        if task_id == 1:
            if x == cost_J:
                prob_intersection += joint_dist.pmf(x, y)
        elif task_id == 2:
            if y == cost_J:
                prob_intersection += joint_dist.pmf(x, y)
        else:
            raise ValueError("Invalid task_id; must be either 1 or 2.")

    # Compute the conditional probability
    if prob_event == 0:
        return 0  # The conditional probability is undefined if P(event) = 0, but we return 0 for computational purposes
    else:
        conditional_prob = prob_intersection / prob_event

    return conditional_prob
    
def is_conditionally_independent(task_id, cost, cost_other, event, joint_dist):
    """
    Check if two random variables are conditionally independent given an event.

    Parameters:
    - task_id: int
        The task ID, which should be either 1 or 2. Determines the role of cost and cost_other.
    - cost: numeric
        The cost value for task J.
    - cost_other: numeric
        The cost value for the other task.
    - event: set of tuples
        A set of (x, y) tuples representing the event in the sample space.
    - joint_dist: JointRvDiscrete object
        Object representing the joint distribution.

    Returns:
    - bool
        True if the variables are conditionally independent, False otherwise.
    """
    
    # Compute the joint conditional probability P((cost, cost_other) | event)
    joint_cond_prob = conditional_probability_cost_cost_other_event(task_id, cost, cost_other, event, joint_dist)
    
    # Compute the individual conditional probabilities P(cost | event)
    cond_prob_cost = conditional_probability_cost_event(task_id, cost, event, joint_dist)

    # Determine the task_id for the "other" task
    other_task_id = 2 if task_id == 1 else 1
    
    # Compute the individual conditional probabilities P(cost_other | event)
    cond_prob_cost_other = conditional_probability_cost_event(other_task_id, cost_other, event, joint_dist)

    # Check for conditional independence
    return np.isclose(joint_cond_prob, cond_prob_cost * cond_prob_cost_other)


def is_conditionally_independent_all_costs(task_id, event, joint_dist):
    """
    Check if two random variables are conditionally independent given an event for all possible cost values.

    Parameters:
    - task_id: int
        The task ID, which should be either 1 or 2. Determines the role of cost and cost_other.
    - event: set of tuples
        A set of (x, y) tuples representing the event in the sample space.
    - joint_dist: JointRvDiscrete object
        Object representing the joint distribution.

    Returns:
    - bool
        True if the variables are conditionally independent for all cost values, False otherwise.
    """
    
    # Error handling for invalid task_id
    if task_id not in [1, 2]:
        raise ValueError("Invalid task_id; must be either 1 or 2.")
    
    # Determine the task_id for the "other" task
    other_task_id = 2 if task_id == 1 else 1
    
    # Loop through all possible cost values for the task specified by task_id
    for cost in joint_dist.x_values if task_id == 1 else joint_dist.y_values:
        
        # Loop through all possible cost values for the "other" task
        for cost_other in joint_dist.x_values if other_task_id == 1 else joint_dist.y_values:
            
            # Check if the variables are conditionally independent given the event
            if not is_conditionally_independent(task_id, cost, cost_other, event, joint_dist):
                return False
                
    return True


def is_conditionally_independent_partition(task_id, partition, joint_dist):
    """
    Checks if all events in a partition are conditionally independent with respect to a given joint distribution.
    
    Parameters:
    - task_id (Any): An identifier for the task.
    - partition (List[Event]): A list of events in the partition.
    - joint_dist (JointDistribution): The joint distribution of interest.
    
    Returns:
    - bool: True if all events in the partition are conditionally independent, False otherwise.
    """
    
    for event in partition:
        if not is_conditionally_independent_all_costs(task_id, event, joint_dist):
            return False
    
    return True

def filter_independent_partitions(task_id, joint_dist):
    """
    Filters partitions based on conditional independence criteria.
    
    Args:
        task_id: An identifier for the task. The type depends on how it's used in `is_conditionally_independent_partition`.
        joint_dist: The joint distribution data structure. Expected to have attributes `x_values` and `y_values`.
    
    Returns:
        List: A list of partitions that satisfy the conditional independence criteria.
    """
    
    # Generate the sample space omega as a Cartesian product of x_values and y_values
    omega = set(product(joint_dist.x_values, joint_dist.y_values))
    
    # Generate all possible partitions of omega
    partitions = generate_partitions(omega)
    
    # Initialize an empty list to hold partitions that meet the conditional independence criteria
    independent_partitions = []
    
    # Iterate over each partition to check for conditional independence
    for partition in partitions:
        
        # Use is_conditionally_independent_partition to verify if the current partition meets the criteria
        if is_conditionally_independent_partition(task_id, partition, joint_dist):
            
            # If it does, append it to the independent_partitions list
            independent_partitions.append(partition)
    
    # Return the list of partitions that satisfy the conditional independence criteria
    return independent_partitions

from scipy.stats import rv_discrete

def derive_pET(task_id, event, joint_dist):
    """
    Derive the Probabilistic Execution Time distribution (pET) for a given task based on a specified event and provided joint distribution.

    Parameters:
    task_id (int): ID of the task for which pET needs to be derived.
    event (list of tuple): List of tuples defining a specific event in the system.
    joint_dist (object): Object representing the joint distribution of costs for two tasks.
    
    Returns:
    rv_discrete: A random variable object representing the pET distribution.
    """
    pET_values = []
    pET_probabilities = []
    for cost in joint_dist.x_values if task_id == 1 else joint_dist.y_values:
        prob_cost = conditional_probability_cost_event(task_id, cost, event, joint_dist)
        pET_values.append(cost)
        pET_probabilities.append(prob_cost)
    return rv_discrete(values=(pET_values, pET_probabilities))

def derive_pWCET(task_id, partition, joint_dist):
    """
    Derive the pWCET (probabilistic Worst-Case Execution Time) distribution for a given task and partition.
    
    This function computes the pWCET distribution by taking the pointwise minimum of the CDFs (Cumulative 
    Distribution Functions) of the pET (probabilistic Execution Time) distributions corresponding to each 
    event in the partition. The resulting distribution is then converted to its non-zero form and returned.

    Parameters:
    - task_id (int): The ID of the task for which the pWCET is to be derived.
    - partition (list): A list of events constituting the partition.
    - joint_dist (JointRvDiscrete object): The joint distribution of execution times for tasks.
    
    Returns:
    - rv_discrete: The derived pWCET as a rv_discrete object.
    """
    min_CDF_RV = rv_discrete(values=([0], [1]))
    for event in partition:
        min_CDF_RV = pointwise_min_cdfs([min_CDF_RV, derive_pET(task_id, event, joint_dist)])
    min_CDF_RV_non_zero = to_non_zero_rv(min_CDF_RV)
    min_CDF_RV = to_rv_discrete(min_CDF_RV_non_zero[0], min_CDF_RV_non_zero[1])
    return min_CDF_RV

def is_same_distribution(rv1, rv2):
    """
    Check if two rv_discrete objects represent the same probability distribution.
    
    This function compares both the values (xk) and the probabilities (pk) of two given
    rv_discrete objects to determine if they represent the same probability distribution.
    
    Parameters:
    - rv1 (rv_discrete): The first probability distribution as an rv_discrete object.
    - rv2 (rv_discrete): The second probability distribution as an rv_discrete object.
    
    Returns:
    - bool: True if both rv_discrete objects represent the same distribution; False otherwise.
    """
    return np.array_equal(rv1.xk, rv2.xk) and np.array_equal(rv1.pk, rv2.pk)

def pretty_print_rv_discrete(rv, name):
    """
    Pretty print an rv_discrete object with aligned values and probabilities.

    Parameters:
    - rv (rv_discrete): The rv_discrete object containing the distribution to be printed.
    - name (str): An arbitrary name provided before printing the distribution.
    """
    print(name)
    value_str = " ".join([f"{x: >5}" for x in rv.xk])
    prob_str = " ".join([f"{p: >5.2f}" for p in rv.pk])
    print(f"Values:      {value_str}")
    print(f"Probabilities:{prob_str}")

def derive_pWCETs_all_partitions(task_id, joint_dist, print_debug=False):
    """
    Derive the probabilistic Worst-Case Execution Time (pWCET) distributions for all conditionally-independent partitions.

    This function iterates over all conditionally-independent partitions for a given task and derives their respective
    pWCET distributions. It avoids appending duplicates to the final list of pWCET distributions.

    Parameters:
    - task_id (int): The ID of the task for which the pWCETs are being derived.
    - joint_dist (JointRvDiscrete): The joint distribution of execution times for two tasks as a JointRvDiscrete object.

    Returns:
    - list[rv_discrete]: A list of unique pWCET distributions as rv_discrete objects.

    """
    partitions = filter_independent_partitions(task_id, joint_dist)
    pWCET_distributions = []

    for partition in partitions:
        pWCET = derive_pWCET(task_id, partition, joint_dist)
        if print_debug:
            print("\nPartition:", partition)
            pretty_print_rv_discrete(pWCET, "pWCET")
        # Check if pWCET already exists in pWCET_distributions
        exists = any(is_same_distribution(pWCET, existing_pWCET) for existing_pWCET in pWCET_distributions)
        
        if not exists:
            pWCET_distributions.append(pWCET)

    return pWCET_distributions

def derive_minimised_DFP_with_pWCETs(joint_dist, D, print_debug=False):
    """
    Derive the minimised Deadline Failure Probability (DFP) considering all pairs of pWCETs for two tasks.

    This function takes the joint distribution of execution times for two tasks and a deadline D, 
    and computes the minimised DFP. It iterates over all pairs of pWCETs (probabilistic Worst-Case 
    Execution Time) distributions for each task to compute the corresponding DFP values. It returns 
    the minimum DFP found across all such pairs.

    Parameters:
    - joint_dist (JointRvDiscrete): The joint distribution of execution times for two tasks, represented 
                                    as a JointRvDiscrete object.
    - D (int or float): The deadline for task completion.

    Returns:
    - float: The minimum DFP across all considered pairs of pWCET distributions.

    """
    pWCETs_1 = derive_pWCETs_all_partitions(1, joint_dist)
    pWCETs_2 = derive_pWCETs_all_partitions(2, joint_dist)

    min_DFP = 1
    for pWCET_1 in pWCETs_1:
        for pWCET_2 in pWCETs_2:
            pWCET_sum = fast_sum_multiple_rvs([pWCET_1, pWCET_2],[1, 1])
            DFP = compute_independence_assuming_DFP(pWCET_sum, D)
            if print_debug:
                pretty_print_rv_discrete(pWCET_1, "pWCET_1")
                pretty_print_rv_discrete(pWCET_2, "pWCET_2")
                pretty_print_rv_discrete(pWCET_sum, "pWCET_sum")
                print("DFP:", DFP)
            if DFP < min_DFP:
                min_DFP = DFP
    return min_DFP

def derive_CTA_DFP(joint_dist,D, print_debug=False):
    """
    Compute the Deadline Failure Probability (DFP) using Correlation-Tolerant Analysis (CTA).
    
    This function calculates the DFP based on CTA, which aims to account for the statistical 
    correlation between tasks. It computes the expectation and standard deviation for each task's
    execution time, and then derives the DFP using these parameters and the given deadline D.

    Parameters:
    - joint_dist (JointRvDiscrete): The joint distribution of execution times for two tasks, represented 
                                    as a JointRvDiscrete object.
    - D (int or float): The deadline for task completion.

    Returns:
    - float: The computed DFP based on Correlation-Tolerant Analysis.
    
    """
    C_1 = joint_dist.marginal_X
    C_2 = joint_dist.marginal_Y
    E_C_1 = C_1.expect()
    E_C_2 = C_2.expect()
    sd_C_1 = C_1.std()
    sd_C_2 = C_2.std()

    sdSum = sd_C_1 + sd_C_2
    eSum = E_C_1 + E_C_2

    DFP = sdSum**2 / ( sdSum**2 + ( D - (eSum))**2 )

    if print_debug:
        print(f"Expected value of task 1       : {E_C_1:10.6f}")
        print(f"Expected value of task 2       : {E_C_2:10.6f}")
        print(f"Standard deviation of task 1   : {sd_C_1:10.6f}")
        print(f"Standard deviation of task 2   : {sd_C_2:10.6f}")
        print(f"Deadline First Policy (DFP)    : {DFP:10.6f}")

    return DFP

