import unittest
import joint_distribution as jd
import numpy as np
from probability_operations.convolution import fast_sum_multiple_rvs
from scipy.stats import rv_discrete

class TestJointRvDiscrete(unittest.TestCase):

    def setUp(self):
        self.C_1_values = np.array([1, 3])
        self.C_2_values = np.array([2, 8])
        self.joint_pmf = np.array([[0.25, 0.25], [0.25, 0.25]])
        self.joint_rv = jd.JointRvDiscrete(self.joint_pmf, self.C_1_values, self.C_2_values)

    def test_pmf(self):
        self.assertEqual(self.joint_rv.pmf(1, 1), 0)  
        self.assertEqual(self.joint_rv.pmf(1, 2), 0.25)

    def test_compute_ground_truth_DFP(self):
        D = 6
        actual_DFP = jd.compute_ground_truth_DFP(self.joint_rv, D)
        self.assertEqual(actual_DFP, 0.5)

        D = 11
        actual_DFP = jd.compute_ground_truth_DFP(self.joint_rv, D)
        self.assertEqual(actual_DFP, 0)

        D = 10
        actual_DFP = jd.compute_ground_truth_DFP(self.joint_rv, D)
        self.assertEqual(actual_DFP, 0.25)
    
    def test_compute_independence_assuming_DFP(self):

        C1 = self.joint_rv.marginal_X # Random variable representing execution-time distribution of task 1
        C2 = self.joint_rv.marginal_Y # Random variable representing execution-time distribution of task 2
        response_time_RV = fast_sum_multiple_rvs([C1, C2],[1, 1])

        D = 6
        actual_DFP = jd.compute_independence_assuming_DFP(response_time_RV, D)
        self.assertAlmostEqual(actual_DFP, 0.5)

        D = 11
        actual_DFP = jd.compute_independence_assuming_DFP(response_time_RV, D)
        self.assertAlmostEqual(actual_DFP, 0)

        D = 10
        actual_DFP = jd.compute_independence_assuming_DFP(response_time_RV, D)
        self.assertAlmostEqual(actual_DFP, 0.25)

    def test_probability_of_event(self):
        event = [(1,2),(1,8)]
        actual_prob = jd.probability_of_event(event, self.joint_rv)
        self.assertEqual(actual_prob, 0.5)

        event = [(1,2),(1,8),(3,2)]
        actual_prob = jd.probability_of_event(event, self.joint_rv)
        self.assertEqual(actual_prob, 0.75)

        event = [(4,4)]
        actual_prob = jd.probability_of_event(event, self.joint_rv)
        self.assertEqual(actual_prob, 0)
    
    def test_conditional_probability_cost_cost_other_event(self):
        task_id = 1 # Task 1
        cost_J = 1
        cost_J_other = 2

        event = [(1,2),(1,8)]
        actual_prob = jd.conditional_probability_cost_cost_other_event(task_id, cost_J, cost_J_other, event, self.joint_rv)
        self.assertEqual(actual_prob, 0.5)

        event = [(1,2),(1,8),(3,2)]
        actual_prob = jd.conditional_probability_cost_cost_other_event(task_id, cost_J, cost_J_other, event, self.joint_rv)
        self.assertAlmostEqual(actual_prob, 0.3333333333333333)

        event = [(1,8)]
        actual_prob = jd.conditional_probability_cost_cost_other_event(task_id, cost_J, cost_J_other, event, self.joint_rv)
        self.assertAlmostEqual(actual_prob, 0)
        
        task_id = 2 # Task 2
        cost_J = 2
        cost_J_other = 1
        event = [(1,2)]
        actual_prob = jd.conditional_probability_cost_cost_other_event(2, cost_J, cost_J_other, event, self.joint_rv)
        self.assertAlmostEqual(actual_prob, 1)

    def test_conditional_probability_cost_event(self):
        task_id = 1 # Task 1
        cost_J = 1
        
        event = [(1,2),(1,8)]
        actual_prob = jd.conditional_probability_cost_event(task_id, cost_J, event, self.joint_rv)
        self.assertEqual(actual_prob, 1)
        
        event = [(1,2),(1,8),(3,2)]
        actual_prob = jd.conditional_probability_cost_event(task_id, cost_J, event, self.joint_rv)
        self.assertAlmostEqual(actual_prob, 0.6666666666666666)

        task_id = 2 # Task 2
        cost_J = 2
        event = [(1,2),(1,8)]
        actual_prob = jd.conditional_probability_cost_event(task_id, cost_J, event, self.joint_rv)
        self.assertEqual(actual_prob, 0.5)

    def test_is_conditionally_independent(self):
        task_id = 1
        cost = 1
        cost_other = 2

        event = [(1,2),(1,8)]
        actual_result = jd.is_conditionally_independent(task_id, cost, cost_other, event, self.joint_rv)
        self.assertEqual(actual_result, True)

        event = [(1,2),(1,8),(3,2)]
        actual_result = jd.is_conditionally_independent(task_id, cost, cost_other, event, self.joint_rv)
        self.assertEqual(actual_result, False)

        event = [(1,2),(1,8),(3,2),(3,8)]
        actual_result = jd.is_conditionally_independent(task_id, cost, cost_other, event, self.joint_rv)
        self.assertEqual(actual_result, True)

        task_id = 2
        cost = 2
        cost_other = 1

        event = [(1,2),(1,8),(3,2)]
        # P(C_2 = 2 , C_1 = 1 | event) = 0.25/0.75 = 1/3
        # P(C_2 = 2 | event) = 0.25/0.75 + 0.25/0.75 = 2/3
        # P(C_1 = 1 | event) = 0.25/0.75 + 0.25/0.75 = 2/3
        # P(C_2 = 2 | event) * P(C_1 = 1 | event) = 2/3 * 2/3 = 4/9
        # Not independent since 1/3 != 4/9
        actual_result = jd.is_conditionally_independent(task_id, cost, cost_other, event, self.joint_rv)
        self.assertEqual(actual_result, False)

        event = [(1,2),(1,8)]
        # P(C_2 = 2 , C_1 = 1 | event) = 0.25/0.5 = 1/2
        # P(C_2 = 2 | event) = 0.25/0.5 = 1/2
        # P(C_1 = 1 | event) = 0.25/0.5 + 0.25/0.5 = 1
        # P(C_2 = 2 | event) * P(C_1 = 1 | event) = 1/2 * 1 = 1/2
        # Independent since 1/2 = 1/2
        actual_result = jd.is_conditionally_independent(task_id, cost, cost_other, event, self.joint_rv)
        self.assertEqual(actual_result, True)

    def test_is_conditionally_independent_all_costs(self):
        task_id = 1
        event = [(1,2),(1,8)]
        actual_result = jd.is_conditionally_independent_all_costs(task_id, event, self.joint_rv)
        self.assertEqual(actual_result, True)

        task_id = 2
        event = [(1,2),(1,8)]
        actual_result = jd.is_conditionally_independent_all_costs(task_id, event, self.joint_rv)
        self.assertEqual(actual_result, True)

        task_id = 1
        event = [(1,2),(1,8),(3,2)]
        actual_result = jd.is_conditionally_independent_all_costs(task_id, event, self.joint_rv)
        self.assertEqual(actual_result, False)

        task_id = 2
        event = [(1,2),(1,8),(3,2)]
        actual_result = jd.is_conditionally_independent_all_costs(task_id, event, self.joint_rv)
        self.assertEqual(actual_result, False)

        task_id = 1
        event = [(1,2),(3,8)]
        actual_result = jd.is_conditionally_independent_all_costs(task_id, event, self.joint_rv)
        self.assertEqual(actual_result, False)

        task_id = 1
        event = [(3,2),(1,8)]
        actual_result = jd.is_conditionally_independent_all_costs(task_id, event, self.joint_rv)
        self.assertEqual(actual_result, False)


    def test_partition_generation(self):
        omega = set(jd.product(self.joint_rv.x_values, self.joint_rv.y_values))
        partitions_gen = jd.generate_partitions(omega)
        partitions = []
        for partition in partitions_gen:
            partitions.append(partition)

        expected_partitions =   [[[(1, 8), (3, 2), (3, 8), (1, 2)]],
                                [[(1, 8)], [(3, 2), (3, 8), (1, 2)]],
                                [[(1, 8), (3, 2)], [(3, 8), (1, 2)]],
                                [[(3, 2)], [(1, 8), (3, 8), (1, 2)]],
                                [[(1, 8)], [(3, 2)], [(3, 8), (1, 2)]],
                                [[(1, 8), (3, 2), (3, 8)], [(1, 2)]],
                                [[(3, 2), (3, 8)], [(1, 8), (1, 2)]],
                                [[(1, 8)], [(3, 2), (3, 8)], [(1, 2)]],
                                [[(1, 8), (3, 8)], [(3, 2), (1, 2)]],
                                [[(3, 8)], [(1, 8), (3, 2), (1, 2)]],
                                [[(1, 8)], [(3, 8)], [(3, 2), (1, 2)]],
                                [[(1, 8), (3, 2)], [(3, 8)], [(1, 2)]],
                                [[(3, 2)], [(1, 8), (3, 8)], [(1, 2)]],
                                [[(3, 2)], [(3, 8)], [(1, 8), (1, 2)]],
                                [[(1, 8)], [(3, 2)], [(3, 8)], [(1, 2)]]]

        self.assertEqual(len(partitions), len(expected_partitions))
        for partition in partitions:
            self.assertTrue(partition in expected_partitions)
        for expected_partition in expected_partitions:
            self.assertTrue(expected_partition in partitions)

    def test_filter_independent_partitions(self):
        task_id = 1
        independent_partitions = jd.filter_independent_partitions(task_id, self.joint_rv)
        self.assertEqual(len(independent_partitions), 8)

        expected_partitions =   [[[(1, 8), (3, 2), (3, 8), (1, 2)]],
                                [[(3, 2), (3, 8)], [(1, 8), (1, 2)]],
                                [[(1, 8)], [(3, 2), (3, 8)], [(1, 2)]],
                                [[(1, 8), (3, 8)], [(3, 2), (1, 2)]],
                                [[(1, 8)], [(3, 8)], [(3, 2), (1, 2)]],
                                [[(3, 2)], [(1, 8), (3, 8)], [(1, 2)]],
                                [[(3, 2)], [(3, 8)], [(1, 8), (1, 2)]],
                                [[(1, 8)], [(3, 2)], [(3, 8)], [(1, 2)]]]
        # Partitions with elements of exactly three evolutions are not independent
        # Partitions containing [(1,2),(3,8)] or [(3,2),(1,8)] are not independent

        self.assertEqual(len(independent_partitions), len(expected_partitions))
        for partition in independent_partitions:
            self.assertTrue(partition in expected_partitions)
        for expected_partition in expected_partitions:
            self.assertTrue(expected_partition in independent_partitions)

    def test_derive_pET(self):
        task_id = 1     
        event = [(1, 2), (1, 8)]
        
        # Call the function to test
        actual_pET = jd.derive_pET(task_id, event, self.joint_rv)
        # Extract values and probabilities from the returned object
        values_pET = actual_pET.xk
        probabilities_pET = actual_pET.pk

        np.testing.assert_array_almost_equal(values_pET, np.array([1, 3]), decimal=10)
        np.testing.assert_array_almost_equal(probabilities_pET, np.array([1., 0.]), decimal=10)

        task_id = 2     
        event = [(1, 2), (1, 8)]
        
        # Call the function to test
        actual_pET = jd.derive_pET(task_id, event, self.joint_rv)
        # Extract values and probabilities from the returned object
        values_pET = actual_pET.xk
        probabilities_pET = actual_pET.pk

        np.testing.assert_array_almost_equal(values_pET, np.array([2, 8]), decimal=10)
        np.testing.assert_array_almost_equal(probabilities_pET, np.array([0.5, 0.5]), decimal=10)

        task_id = 2     
        event = [(1, 8), (3, 2), (3, 8), (1, 2)]
        
        # Call the function to test
        actual_pET = jd.derive_pET(task_id, event, self.joint_rv)
        # Extract values and probabilities from the returned object
        values_pET = actual_pET.xk
        probabilities_pET = actual_pET.pk

        np.testing.assert_array_almost_equal(values_pET, np.array([2, 8]), decimal=10)
        np.testing.assert_array_almost_equal(probabilities_pET, np.array([0.5, 0.5]), decimal=10)

    def test_derive_pWCET(self):
        task_id = 1
        partition = [[(1, 8), (3, 2), (3, 8), (1, 2)]]
        pWCET = jd.derive_pWCET(task_id, partition, self.joint_rv)

        np.testing.assert_array_almost_equal(pWCET.xk, np.array([1, 3]), decimal=10)
        np.testing.assert_array_almost_equal(pWCET.pk, np.array([0.5, 0.5]), decimal=10)

        task_id = 2
        partition = [[(1, 8), (3, 2), (3, 8), (1, 2)]]
        pWCET = jd.derive_pWCET(task_id, partition, self.joint_rv)
        np.testing.assert_array_almost_equal(pWCET.xk, np.array([2, 8]), decimal=10)
        np.testing.assert_array_almost_equal(pWCET.pk, np.array([0.5, 0.5]), decimal=10)

        task_id = 1
        partition = [[(1, 8)], [(3, 2)], [(3, 8)], [(1, 2)]]
        pWCET = jd.derive_pWCET(task_id, partition, self.joint_rv)
        np.testing.assert_array_almost_equal(pWCET.xk, np.array([3]), decimal=10)
        np.testing.assert_array_almost_equal(pWCET.pk, np.array([1.]), decimal=10)

        task_id = 2
        partition = [[(1, 8)], [(3, 2)], [(3, 8)], [(1, 2)]]
        pWCET = jd.derive_pWCET(task_id, partition, self.joint_rv)
        np.testing.assert_array_almost_equal(pWCET.xk, np.array([8]), decimal=10)
        np.testing.assert_array_almost_equal(pWCET.pk, np.array([1.]), decimal=10)

        task_id = 2
        partition = [[(1, 8), (3, 8)], [(3, 2), (1, 2)]]
        pWCET = jd.derive_pWCET(task_id, partition, self.joint_rv)
        np.testing.assert_array_almost_equal(pWCET.xk, np.array([8]), decimal=10)
        np.testing.assert_array_almost_equal(pWCET.pk, np.array([1.]), decimal=10)

        task_id = 1
        partition = [[(1, 8), (3, 8)], [(3, 2), (1, 2)]]
        pWCET = jd.derive_pWCET(task_id, partition, self.joint_rv)
        np.testing.assert_array_almost_equal(pWCET.xk, np.array([1, 3]), decimal=10)
        np.testing.assert_array_almost_equal(pWCET.pk, np.array([0.5, 0.5]), decimal=10)

    def test_is_same_distribution(self):
        X = rv_discrete(values=(np.array([1, 2]), np.array([0.5, 0.5])))
        Y = rv_discrete(values=(np.array([1, 2]), np.array([0.5, 0.5])))
        self.assertTrue(jd.is_same_distribution(X, Y))

        X = rv_discrete(values=(np.array([1, 2]), np.array([0.4, 0.6])))
        Y = rv_discrete(values=(np.array([1, 2]), np.array([0.5, 0.5])))
        self.assertFalse(jd.is_same_distribution(X, Y))

    def test_derive_all_pWCETs(self):
        # Test for task_id = 1
        task_id = 1
        expected_pWCETs_1 = [
            {'xk': [1, 3], 'pk': [0.5, 0.5]},
            {'xk': [3], 'pk': [1.0]}
        ]
        
        actual_pWCETs_1 = jd.derive_pWCETs_all_partitions(task_id, self.joint_rv)
        for i, pWCET in enumerate(actual_pWCETs_1):
            self.assertListEqual(list(pWCET.xk), expected_pWCETs_1[i]['xk'])
            self.assertListEqual(list(pWCET.pk), expected_pWCETs_1[i]['pk'])

        # Test for task_id = 2
        task_id = 2
        expected_pWCETs_2 = [
            {'xk': [2, 8], 'pk': [0.5, 0.5]},
            {'xk': [8], 'pk': [1.0]}
        ]
        
        actual_pWCETs_2 = jd.derive_pWCETs_all_partitions(task_id, self.joint_rv)
        for i, pWCET in enumerate(actual_pWCETs_2):
            self.assertListEqual(list(pWCET.xk), expected_pWCETs_2[i]['xk'])
            self.assertListEqual(list(pWCET.pk), expected_pWCETs_2[i]['pk'])


    def test_derive_minimised_DFP_with_pWCETs(self):
        D = 10
        min_DFP = jd.derive_minimised_DFP_with_pWCETs(self.joint_rv, D)
        self.assertAlmostEqual(min_DFP, 0.25)

        D = 11
        min_DFP = jd.derive_minimised_DFP_with_pWCETs(self.joint_rv, D)
        self.assertAlmostEqual(min_DFP, 0.)

        D = 9
        min_DFP = jd.derive_minimised_DFP_with_pWCETs(self.joint_rv, D)
        self.assertAlmostEqual(min_DFP, 0.25)

        D = 8
        min_DFP = jd.derive_minimised_DFP_with_pWCETs(self.joint_rv, D)
        self.assertAlmostEqual(min_DFP, 0.5)

    def test_derive_CTA_DFP(self):
        D = 10
        DFP = jd.derive_CTA_DFP(self.joint_rv, D)
        self.assertAlmostEqual(DFP, 0.64)


# Other test methods

if __name__ == "__main__":
    unittest.main()
