import numpy as np
from scipy.stats import rv_discrete
from collections import defaultdict

def to_rv_discrete(values, probabilities):
    """ Converts a given set of values and probabilities to an rv_discrete object."""
    # Create an rv_discrete object with given values and probabilities
    integer_values = values.astype(int)
    custom_dist = rv_discrete(name='custom', values=(integer_values, probabilities.astype(np.float64)))
    return custom_dist

def to_non_zero_rv(X):
    """ Converts a given rv_discrete object to a 2D numpy array with non-zero probabilities and
     the corresponding values."""
    # Find the indices of non-zero probabilities
    non_zero_indices = X.pk > 0

    # Filter the values and probabilities using the non_zero_indices
    non_zero_values = X.xk[non_zero_indices].astype(int)
    non_zero_probs = X.pk[non_zero_indices].astype(np.float64)

    # Return a 2D numpy array with non-zero values and probabilities
    non_zero_rv = np.array([non_zero_values, non_zero_probs], dtype=np.float64)
    return non_zero_rv


import numpy as np
from scipy.stats import rv_discrete

def pointwise_min_cdfs_inefficient(rvs):
    """ Computes the pointwise minimum of the CDFs of the given rv_discrete objects."""
    # Identify the min and max value with non-zero probability for each rv in rvs
    min_values = [rv.xk.min() for rv in rvs]
    max_values = [rv.xk.max() for rv in rvs]
    
    # Define the range of values of interest
    min_val = min(min_values)
    max_val = max(max_values)
    x_values = np.arange(min_val-1, max_val + 1)
    
    # Derive rv.cdf for the given range of x_values for each rv in rvs
    cdfs = [rv.cdf(x_values) for rv in rvs]
    
    # Perform the minimum on all the x_values
    min_cdf_values = np.amin(cdfs, axis=0)

    # Compute the probability mass function (PMF) from the minimum CDF values
    min_pmf_values = np.diff(min_cdf_values, prepend=0)

    # Construct the rv_discrete object corresponding to the min_cdf_values
    min_rv = rv_discrete(name='pointwise_min', values=(x_values, min_pmf_values))

    return min_rv

def pointwise_min_cdfs(rvs):
    """ Computes the pointwise minimum of the CDFs of the given rv_discrete objects."""

    # First, gather all unique points where the rvs change
    change_points = sorted(set().union(*(set(rv.xk) for rv in rvs)))

    # Compute the CDF for each rv at the change points
    cdfs = np.array([rv.cdf(change_points) for rv in rvs])

    # Calculate the minimum CDF at each change point
    min_cdf_values = np.amin(cdfs, axis=0)

    # Compute the probability mass function (PMF) from the minimum CDF values
    min_pmf_values = np.diff(min_cdf_values, prepend=0)

    # Construct the rv_discrete object corresponding to the min_cdf_values
    min_rv = rv_discrete(name='pointwise_min', values=(change_points, min_pmf_values))

    return min_rv

def construct_rv_total_prob(probabilities, rvs):
    """ Constructs an rv_discrete object that is derived by the law of total probability. """
    # Input validation
    if len(probabilities) != len(rvs):
        raise ValueError("The number of probabilities and RVs must be the same.")
    if not np.isclose(sum(probabilities), 1):
        raise ValueError("The probabilities must sum to 1.")
    if any(p < 0 for p in probabilities):
        raise ValueError("Probabilities must be non-negative.")

    # Create empty defaultdict for values and their probabilities
    values_prob_dict = defaultdict(np.float64)

    # Iterate over RVs
    for prob, rv in zip(probabilities, rvs):
        values, probs = rv.xk, rv.pk

        # Scale probabilities and add to dict
        for v, p in zip(values, probs):
            values_prob_dict[v] += prob * p

    # Create arrays for values and their probabilities
    values, probabilities = zip(*values_prob_dict.items())

    # Normalize probabilities
    probabilities = np.array(probabilities) / sum(probabilities)

    # Create and return new rv_discrete object
    return rv_discrete(name='Z', values=(values, probabilities))


def mgf(t, X):
    """
    Computes the Moment Generating Function (MGF) of a random variable.

    Parameters
    ----------
    t : float
        The point at which to compute the MGF.
    X : scipy.stats.rv_discrete
        The random variable, an instance of rv_discrete.

    Returns
    -------
    float
        The MGF of X evaluated at t.
    """
 
    values = X.xk
    probabilities = X.pk

    exp_values = t * values

    # Check if any value in the exp_values will cause an overflow when used in np.exp()
    if np.any(exp_values > np.log(np.finfo(float).max)):
        return np.inf

    # Compute the MGF using the definition: E[e^(tX)]
    mgf_value = np.sum(np.exp(exp_values) * probabilities)

    return mgf_value

