import numpy as np
from scipy.stats import rv_discrete
from scipy.signal import fftconvolve
from probability_operations.basic_probability_functions import to_non_zero_rv, to_rv_discrete



def sum_rv(X, Y):
    """ Returns the sum of two rv_discrete objects X and Y."""
    # Get non-zero values and probabilities from rv_discrete objects X and Y
    x_values, x_probs = to_non_zero_rv(X)
    y_values, y_probs = to_non_zero_rv(Y)

    # Find the minimum and maximum values in x_values and y_values
    min_value = min(np.min(x_values), np.min(y_values))
    max_value = max(np.max(x_values), np.max(y_values))

    # Calculate the extended size for the convolution
    extended_size = int(max_value + min_value)

    # Create extended probability arrays filled with zeros
    x_extended_probs = np.zeros(extended_size, dtype=np.float64)
    y_extended_probs = np.zeros(extended_size, dtype=np.float64)

    # Calculate indices for x_values and y_values in the extended arrays
    x_indices = (x_values - min_value).astype(int)
    y_indices = (y_values - min_value).astype(int)

    # Fill the extended probability arrays with the corresponding probabilities
    x_extended_probs[x_indices] = x_probs
    y_extended_probs[y_indices] = y_probs

    # Perform the convolution of the two extended probability arrays
    convolved_probs = fftconvolve(x_extended_probs, y_extended_probs, mode='full')
    convolved_values = np.arange(min_value * 2, min_value * 2 + len(convolved_probs), dtype=int)

    # Set any small negative values in convolved_probs to 0
    convolved_probs[convolved_probs < 1e-16] = 0.0

    # Find the non-zero indices in convolved_probs
    non_zero_indices = np.nonzero(convolved_probs)

    # Extract non-zero values and probabilities from convolved arrays
    non_zero_values = convolved_values[non_zero_indices]
    non_zero_probs = convolved_probs[non_zero_indices]

    # Create an rv_discrete object for the sum of X and Y
    Z = to_rv_discrete(non_zero_values, non_zero_probs)
    return Z



def sum_multiple_rvs(rvs, counts):
    """ Returns the sum of multiple rv_discrete objects.
    counts is a list of the number of times each rv_discrete object is repeated in the sum."""
    # Check if the inputs are valid
    if len(rvs) != len(counts):
        raise ValueError("The number of RVs and counts must be the same.")
    
    # Create a list of RVs, each repeated the number of times specified by the count
    repeated_rvs = []
    for rv, count in zip(rvs, counts):
        repeated_rvs.extend([rv] * count)
    
    # If the repeated_rvs list is empty, return an empty list
    if not repeated_rvs:
        return []
    
    # Initialize the result with the first RV
    result = repeated_rvs[0]
    
    # Iterate through the rest of the repeated_rvs list, summing RVs
    for i in range(1, len(repeated_rvs)):
        result = sum_rv(result, repeated_rvs[i])
    
    return result

def fast_sum_rv(X, n):
    """ Returns the sum of an rv_discrete object X, repeated n times.
    Function uses a divide-and-conquer approach to reduce the number of convolutions."""
    # If n is 0, return None since we don't have any repetitions of the RV
    if n == 0:
        return None
    # If n is 1, return the RV itself, as there's no need to sum it with anything
    elif n == 1:
        return X
    # If n is even, divide n by 2 and recursively call fast_sum_rv with the new value
    # Then, sum the result with itself to get the final result
    elif n % 2 == 0:
        half_sum = fast_sum_rv(X, n // 2)
        return sum_rv(half_sum, half_sum)
    # If n is odd, subtract 1 from n and divide the result by 2
    # Then, recursively call fast_sum_rv with the new value
    # Finally, sum the result with X and itself to get the final result
    else:
        half_sum = fast_sum_rv(X, (n - 1) // 2)
        return sum_rv(X, sum_rv(half_sum, half_sum))


def fast_sum_multiple_rvs(rvs, counts):
    """ Returns the sum of multiple rv_discrete objects.
    counts is a list of the number of times each rv_discrete object is repeated in the sum.
    Function uses a divide-and-conquer approach to reduce the number of convolutions."""
    # Check if the inputs are valid
    if len(rvs) != len(counts):
        raise ValueError("The number of RVs and counts must be the same.")
    
    # If the rvs list is empty, return an empty list
    if not rvs:
        return []
    
    # Initialize the result with the first RV, repeated the specified number of times
    result = fast_sum_rv(rvs[0], counts[0])

    # Iterate through the rest of the rvs list, summing RVs
    for i in range(1, len(rvs)):
        repeated_rv = fast_sum_rv(rvs[i], counts[i])
        if repeated_rv:
            result = sum_rv(result, repeated_rv)

    return result

