import unittest
import numpy as np
from convolution import to_rv_discrete, sum_rv, to_non_zero_rv, sum_multiple_rvs, fast_sum_multiple_rvs, fast_sum_rv
from basic_probability_functions import to_rv_discrete, to_non_zero_rv, pointwise_min_cdfs, construct_rv_total_prob, mgf

class TestBasicProbability(unittest.TestCase):

    def setUp(self):
        # Create rv_discrete objects X
        self.X_values = np.array([1, 3])
        self.X_probs = np.array([0.5, 0.5], dtype=np.float64)
        self.X = to_rv_discrete(self.X_values, self.X_probs)
        # Create rv_discrete objects Y
        self.Y_values = np.array([4, 5])
        self.Y_probs = np.array([0.5, 0.5], dtype=np.float64)
        self.Y = to_rv_discrete(self.Y_values, self.Y_probs)

        # Create rv_discrete objects V
        self.V_values = np.array([1, 30])
        self.V_probs = np.array([0.6, 0.4], dtype=np.float64)
        self.V = to_rv_discrete(self.V_values, self.V_probs)

        # Create rv_discrete objects W
        self.W_values = np.array([8, 11])
        self.W_probs = np.array([0.2, 0.8], dtype=np.float64)
        self.W = to_rv_discrete(self.W_values, self.W_probs)

        # Create rv_discrete objects W2
        self.W2_values = np.array([8, 11])
        self.W2_probs = np.array([0.5, 0.5], dtype=np.float64)
        self.W2 = to_rv_discrete(self.W2_values, self.W2_probs)
    
    def test_to_rv_discrete(self):
        # Define the expected result
        expected_result = np.array([1, 3])
        expected_probs = np.array([0.5, 0.5], dtype=np.float64)

        # Check if the computed result matches the expected result
        np.testing.assert_array_equal(self.X.xk, expected_result)
        np.testing.assert_array_equal(self.X.pk, expected_probs)

    def test_pointwise_min_cdfs(self):

        # Compute the point-wise minimum of the X and Y
        min_cdfs = to_non_zero_rv(pointwise_min_cdfs([self.X, self.Y]))

        # Define the expected result
        expected_result = np.array([
                            [4, 5],
                            [0.5, 0.5]
                            ])
        # Check if the computed result matches the expected result
        np.testing.assert_array_almost_equal(min_cdfs, expected_result, decimal=6)

        # Compute the point-wise minimum of the V, W and Y
        min_cdfs = to_non_zero_rv(pointwise_min_cdfs([self.V, self.W, self.Y]))

        # Define the expected result
        expected_result = np.array([
            [8,   11,  30],
            [0.2, 0.4, 0.4]
        ])
        # Check if the computed result matches the expected result
        np.testing.assert_array_almost_equal(min_cdfs, expected_result, decimal=6)

    def test_construct_rv_total_prob(self):

        #===== TEST for X, Y, W =====#
        prob_list = [0.7, 0.2, 0.1]
        rv_list = [self.X, self.Y, self.W]

        # Call the function with these inputs
        Z = construct_rv_total_prob(prob_list, rv_list)

        # Check the values of the resulting random variable
        expected_values = np.array([1, 3, 4, 5, 8, 11])
        np.testing.assert_array_equal(Z.xk, expected_values)

        # Check the probabilities of the resulting random variable
        expected_probs = np.array([0.7*0.5, 0.7*0.5, 0.2*0.5, 0.2*0.5, 0.1*0.2, 0.1*0.8])
        np.testing.assert_array_almost_equal(Z.pk, expected_probs, decimal=6)
        
        #===== TEST for W and W2 =====#
        prob_list = [0.6,0.4]
        rv_list = [self.W, self.W2]

        # Call the function with these inputs
        Z = construct_rv_total_prob(prob_list, rv_list)

        # Check the values of the resulting random variable
        expected_values = np.array([8, 11])
        np.testing.assert_array_equal(Z.xk, expected_values)

        # Check the probabilities of the resulting random variable
        expected_probs = np.array([0.6*0.2+0.4*0.5, 0.6*0.8+0.4*0.5])
        np.testing.assert_array_almost_equal(Z.pk, expected_probs, decimal=6)


        #===== TEST for X, Y, W, and W2 =====#
        prob_list = [0.4, 0.3, 0.2, 0.1]
        rv_list = [self.X, self.Y, self.W, self.W2]

        # Call the function with these inputs
        Z = construct_rv_total_prob(prob_list, rv_list)

        # Check the values of the resulting random variable
        expected_values = np.array([1, 3, 4, 5, 8, 11])
        np.testing.assert_array_equal(Z.xk, expected_values)

        # Check the probabilities of the resulting random variable
        expected_probs = np.array([0.4*0.5, 0.4*0.5, 0.3*0.5, 0.3*0.5, 0.2*0.2+0.1*0.5, 0.2*0.8+0.1*0.5])
        np.testing.assert_array_almost_equal(Z.pk, expected_probs, decimal=6)
        

    def test_mgf(self):
        t = 1
        expected_result = 0.5*np.exp(t * 1) + 0.5*np.exp(t*3)
        self.assertAlmostEqual(mgf(t,self.X), expected_result, places=6)

        t = 10
        expected_result = 0.5*np.exp(t * 1) + 0.5*np.exp(t*3)
        self.assertAlmostEqual(mgf(t,self.X), expected_result, places=6)

if __name__ == '__main__':
    unittest.main()