import unittest
import numpy as np
from convolution import to_rv_discrete, sum_rv, to_non_zero_rv, sum_multiple_rvs, fast_sum_multiple_rvs, fast_sum_rv

class TestConvolution(unittest.TestCase):

    def setUp(self):
        # Create rv_discrete objects X
        self.X_values = np.array([1, 3])
        self.X_probs = np.array([0.5, 0.5], dtype=np.float64)
        self.X = to_rv_discrete(self.X_values, self.X_probs)
        # Create rv_discrete objects Y
        self.Y_values = np.array([4, 5])
        self.Y_probs = np.array([0.5, 0.5], dtype=np.float64)
        self.Y = to_rv_discrete(self.Y_values, self.Y_probs)

        # Create rv_discrete objects V
        self.V_values = np.array([1, 30])
        self.V_probs = np.array([0.6, 0.4], dtype=np.float64)
        self.V = to_rv_discrete(self.V_values, self.V_probs)
        # Create rv_discrete objects W
        self.W_values = np.array([4, 10])
        self.W_probs = np.array([0.5, 0.5], dtype=np.float64)
        self.W = to_rv_discrete(self.W_values, self.W_probs)

    def test_sum_rv(self):
        # Compute the sum of X and Y
        Z = sum_rv(self.X, self.Y)
        Z_non_zero = to_non_zero_rv(Z)

        expected_values = np.array([5, 6, 7, 8])
        expected_probs = np.array([0.25, 0.25, 0.25, 0.25], dtype=np.float64)

        np.testing.assert_array_equal(Z_non_zero[0], expected_values)
        np.testing.assert_array_almost_equal(Z_non_zero[1], expected_probs, decimal=15)

        # Compute the sum of X and X
        Z = sum_rv(self.X, self.X)
        Z_non_zero = to_non_zero_rv(Z)

        expected_values = np.array([2, 4, 6])
        expected_probs = np.array([0.25, 0.5, 0.25], dtype=np.float64)

        np.testing.assert_array_equal(Z_non_zero[0], expected_values)
        np.testing.assert_array_almost_equal(Z_non_zero[1], expected_probs, decimal=15)

    def test_sum_multiple_rvs(self):
        # Compute the sum of X, X, and X
        Z = sum_multiple_rvs([self.X],[3])
        Z_non_zero = to_non_zero_rv(Z)

        expected_values = np.array([3, 5, 7, 9])
        expected_probs = np.array([0.125, 0.375, 0.375, 0.125], dtype=np.float64)

        np.testing.assert_array_equal(Z_non_zero[0], expected_values)
        np.testing.assert_array_almost_equal(Z_non_zero[1], expected_probs, decimal=15)

    def test_fast_sum_multiple_rvs(self):
        # Compute the fast sum of X, X, and X
        Z = fast_sum_multiple_rvs([self.X],[3])
        Z_non_zero = to_non_zero_rv(Z)

        expected_values = np.array([3, 5, 7, 9])
        expected_probs = np.array([0.125, 0.375, 0.375, 0.125], dtype=np.float64)

        np.testing.assert_array_equal(Z_non_zero[0], expected_values)
        np.testing.assert_array_almost_equal(Z_non_zero[1], expected_probs, decimal=15)

        # Compute the fast sum of X+X+X+Y+Y
        Z = fast_sum_multiple_rvs([self.X, self.Y],[3,2])
        Z_non_zero = to_non_zero_rv(Z)

        expected_values = np.array([11,12,13,14,15,16,17,18,19])
        expected_probs = np.array([0.03125, 0.0625 , 0.125  , 0.1875 , 0.1875 , 0.1875 , 0.125  , 0.0625 , 0.03125], dtype=np.float64)

        np.testing.assert_array_equal(Z_non_zero[0], expected_values)
        np.testing.assert_array_almost_equal(Z_non_zero[1], expected_probs, decimal=15)

if __name__ == '__main__':
    unittest.main()